import logging
from multiprocessing import freeze_support

import uvicorn
from fastapi import Depends, FastAPI
from fastapi.middleware import Middleware
from fastapi.middleware.httpsredirect import HTTPSRedirectMiddleware

import eventlog
from container import Container
from core import LoggedBodyMiddleware
from core.middleware import ExceptionMiddleware, log_request_body
from core.settings import config
from environment.checks import PlatformEnvironment
from handlers.auth import auth_router
from handlers.base import base_router
from handlers.meta import meta_router
from logging_config import LOGGING_CONFIG

logging.config.dictConfig(LOGGING_CONFIG)


def get_middlewares() -> list:
    """Getting the list of middlewares.

    Returns:
        list: List of middlewares.
    """
    ssa_middlewares = []
    # Logging middleware should always be first
    ssa_middlewares.append(Middleware(LoggedBodyMiddleware))
    ssa_middlewares.append(Middleware(ExceptionMiddleware))
    # Redirect all HTTP requests to HTTPS
    if config.SSL.USE_HTTPS:
        ssa_middlewares.append(Middleware(HTTPSRedirectMiddleware))

    return ssa_middlewares


def get_application() -> FastAPI:
    """Make FastAPI application.

    Returns:
        FastAPI: FastAPI application
    """
    application = FastAPI(middleware=get_middlewares(), dependencies=[Depends(log_request_body)])
    application.include_router(base_router)

    # Include router only if platform is Windows
    if PlatformEnvironment.is_windows():
        from handlers.rds import rds_router

        application.include_router(rds_router)

    # Include router only if platform is Linux
    if PlatformEnvironment.is_linux():
        from handlers.stal import stal_router

        application.include_router(stal_router)

    application.include_router(meta_router, prefix="/meta", tags=["Meta"])
    application.include_router(auth_router, tags=["Auth"])

    return application


app = get_application()


def start_agent():
    ssl_private_key = str(config.SSL.PRIVATE_KEY_PATH) if config.SSL.USE_HTTPS else None
    ssl_cert = str(config.SSL.CERTIFICATE_PATH) if config.SSL.USE_HTTPS else None

    uvicorn.run(
        "app:app",
        host=config.AGENT.LISTEN_IP,
        port=config.AGENT.LISTEN_PORT,
        ssl_keyfile=ssl_private_key,
        ssl_certfile=ssl_cert,
    )


if __name__ == "__main__":
    container = Container()
    freeze_support()
    eventlog.write_event("The Sessional Agent starts up.", "INFO")
    config.check_health_token()
    config.configurate_rds_role()
    start_agent()
