import configparser
import logging
import re
from pathlib import Path
from threading import Thread
from typing import Any, Dict

from pydantic import BaseSettings, Extra

from environment.checks import PlatformEnvironment
from scripts.ssl import generate_selfsigned_cert

if PlatformEnvironment.is_linux():
    from constants.linux.constants import ConstLinuxEnv as const

    IS_WINDOWS = False
elif PlatformEnvironment.is_windows():
    from constants.windows.constants import ConstWindowsEnv as const

    IS_WINDOWS = True


logger = logging.getLogger("settings.handler")


class _ConfigHandler:
    @staticmethod
    def get_params(filepath: Path = Path(const.CONFIG_FILE_PATH)) -> Dict[str, Any]:
        if not filepath.exists():
            logger.debug(f"Указан несуществующий файл конфигурации по адресу: {filepath}")

            return {}

        ssa_conf = configparser.ConfigParser()

        # Without this option values in config will be lowercase.
        ssa_conf.optionxform = str
        ssa_conf.read(filepath)

        # Вызывается таким образом, а не через dict(ssa_conf),
        # поскольку тогда возвращается словарь с дополнительным ключом DEFAULT,
        # который нельзя удалить.
        #
        # del ssa_conf["DEFAULT"] -> raise KeyError
        return {s: dict(ssa_conf[s]) for s in ssa_conf.sections()}

    @staticmethod
    def sync_params(params: Dict[str, Any], filepath: Path = Path(const.CONFIG_FILE_PATH)) -> None:
        if not params:
            logger.error("Нет данных для записи конфигурационного файла.")
            return

        cfg = configparser.ConfigParser()
        cfg.optionxform = str  # type: ignore
        cfg.read_dict(params)

        with open(filepath, mode="w", encoding="utf-8") as cfg_f:
            cfg.write(cfg_f)


def ini_file_settings_source(settings: BaseSettings) -> Dict[str, Any]:
    """
    Получение информации о настройках из файла конфигурации
    """

    return _ConfigHandler.get_params()


class AgentSettings(BaseSettings):
    LISTEN_IP: str = const.DEFAULT_LISTEN_IP
    LISTEN_PORT: int = const.DEFAULT_LISTEN_PORT
    HEALTH_CHECK_ACCESS_KEY: str = const.DEFAULT_HEALTHCHECK_TOKEN
    INSTALL_RDS_ROLE_ON_NEXT_BOOT: bool = const.DEFAULT_INSTALL_RDS_ROLE_ON_NEXT_BOOT

    class Config:
        extra = Extra.ignore


class SSLSettings(BaseSettings):
    USE_HTTPS: bool = const.USE_HTTPS
    CERTIFICATE_PATH: Path = const.SESSION_AGENT_CONFIG["SSL"]["CERTIFICATE_PATH"]
    PRIVATE_KEY_PATH: Path = const.SESSION_AGENT_CONFIG["SSL"]["PRIVATE_KEY_PATH"]

    class Config:
        extra = Extra.ignore


class LoggingSettings(BaseSettings):
    LEVEL: str = const.SESSION_AGENT_CONFIG["LOGGING"]["LEVEL"]
    ROTATION_LIMIT: int = const.SESSION_AGENT_CONFIG["LOGGING"]["ROTATION_LIMIT"]
    MAX_FILE_SIZE: int = const.SESSION_AGENT_CONFIG["LOGGING"]["MAX_FILE_SIZE"]

    class Config:
        extra = Extra.ignore


class Settings(BaseSettings):
    AGENT: AgentSettings = AgentSettings()

    SSL: SSLSettings = SSLSettings()

    LOGGING: LoggingSettings = LoggingSettings()

    class Config:
        env_file_encoding = "utf-8"

        @classmethod
        def customise_sources(cls, init_settings, env_settings, file_secret_settings):
            """
            Официальная документация:
            https://docs.pydantic.dev/1.10/usage/settings/#customise-settings-sources

            Метод для модификации источников данных для конфигурации ресурсов
            """

            return (init_settings, ini_file_settings_source, file_secret_settings)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.__update_config_file()

        try:
            self._check_cert_files()
        except FileNotFoundError as ex:
            if not self.is_using_default_cert():
                raise FileNotFoundError(ex) from ex

            generate_selfsigned_cert(key_file=self.SSL.PRIVATE_KEY_PATH, cert_file=self.SSL.CERTIFICATE_PATH)

    def __update_config_file(self) -> None:
        """Update config file SSA"""
        _ConfigHandler.sync_params(self.dict())

    def call_install_role(self) -> None:
        """Install rds role"""

        from utils.meta.rds import WindowsConfigurator

        try:
            win_config = WindowsConfigurator()
            win_config.configurate_all()
            self.AGENT.INSTALL_RDS_ROLE_ON_NEXT_BOOT = False
            self.__update_config_file()
            logger.info("Install rds role is sucuess")

        except Exception:
            logger.warning("Install rds role end with error")

    def configurate_rds_role(self) -> None:
        """Configurate rds role"""

        if IS_WINDOWS:
            from utils.rds import check_instance_is_rds

            is_install_role = self.AGENT.INSTALL_RDS_ROLE_ON_NEXT_BOOT
            if is_install_role and not check_instance_is_rds():
                thread = Thread(target=self.call_install_role)
                thread.start()

    def check_health_token(self):
        """
        Check HEALTH token
        """
        token = self.AGENT.HEALTH_CHECK_ACCESS_KEY
        if token != "None" and token and not re.match(r"[a-f0-9]+", token):
            err_msg = "Token validation warning. The healthCheckAccessKey doesn't match regexp."
            logger.warning(err_msg)

    def is_using_default_cert(self) -> bool:
        """
        Checking if SSL certificate and private key exists and located in default paths.

        This verification is required to verify that
        all conditions are met before generating a
        self-signed certificate.

        Returns:
            bool: The using default sertificate.
        """

        is_default_cert: bool = (
            self.SSL.CERTIFICATE_PATH == const.SESSION_AGENT_CONFIG["SSL"]["CERTIFICATE_PATH"]
        )
        is_default_key: bool = (
            self.SSL.PRIVATE_KEY_PATH == const.SESSION_AGENT_CONFIG["SSL"]["PRIVATE_KEY_PATH"]
        )

        return is_default_cert and is_default_key

    def _check_cert_files(self) -> None:
        """Checking if cert files is existing.

        Raises:
            FileNotFoundError: The certificate file or private key file was not found.
        """

        if not self.cert_exists():
            err_msg = (
                "Файл сертификата не найден. "
                f"Поместите файл в директорию: {self.SSL.CERTIFICATE_PATH}. "
                f"Или укажите путь до сертификата в конфигурационном файле: {const.CONFIG_FILE_PATH}."
            )

            logger.error(err_msg)

            raise FileNotFoundError(err_msg)

        if not self.private_key_exists():
            err_msg = (
                "Файл приватного ключа не найден. "
                f"Поместите файл в директорию: {self.SSL.PRIVATE_KEY_PATH}. "
                f"Или укажите путь до приватного ключа в конфигурационном файле: {const.CONFIG_FILE_PATH}."
            )

            logger.error(err_msg)

            raise FileNotFoundError(err_msg)

    def cert_exists(self) -> bool:
        """Checking if SSL certificate exist.

        Returns:
            bool: Certificate file exists.
        """
        return self.SSL.CERTIFICATE_PATH.exists()

    def private_key_exists(self) -> bool:
        """Checking if SSL private key exist.

        Returns:
            bool: Private key file exists.
        """
        return self.SSL.PRIVATE_KEY_PATH.exists()


config = Settings()
