import os
import platform
import subprocess
from typing import Callable, Dict

from constants.common_const import ConstEnvironment as env_const


class PlatformEnvironment:
    """Environment checks of the current operating system"""

    def __init__(self):
        self.system = platform.uname().system

    def _check_windows_platform(self, command: str) -> bool:
        """Checking if any vm from VM_LIST in output of command. Windows only."""
        try:
            result = os.popen(command).read()
            is_vm = any(vm in result for vm in env_const.SUPPORTED_VMS)
        # Todo (ap): пустой except. Необходимо конкретизировать ошибку. После чего удалить "# noqa E722"
        except Exception:  # noqa E722
            is_vm = False
        return is_vm

    def _check_windows_serial(self) -> bool:
        """Checking if serial number from output of the command == 0:str. Windows only."""
        try:
            result = os.popen(env_const.HARDWARE_SERIAL_NUMBER).read().replace("\n", "").split()
            is_vm = result[1] == "0"
        # Todo (ap): пустой except. Необходимо конкретизировать ошибку. После чего удалить "# noqa E722"
        except Exception:  # noqa E722
            is_vm = False
        return is_vm

    def _check_linux_platform(self, command: str) -> bool:
        """Checking if Linux system is VM from output of the bash command"""
        try:
            result = os.popen(command).read().rstrip().split(": ")[-1]
            is_vm = bool(result)
        # Todo (ap): пустой except. Необходимо конкретизировать ошибку. После чего удалить "# noqa E722"
        except Exception:  # noqa E722
            is_vm = False
        return is_vm

    def _detect_linux_virt(self) -> bool:
        """Checking if Linux system is VM from output of the bash command"""
        try:
            output = subprocess.run(env_const.LINUX_DETECT_VIRT, stdout=subprocess.PIPE, check=False)
            result = output.stdout.decode().rstrip()
            is_vm = result != env_const.LINUX_DETECT_VIRT_IS_PC
        # Todo (ap): пустой except. Необходимо конкретизировать ошибку. После чего удалить "# noqa E722"
        except Exception:  # noqa E722
            is_vm = False
        return is_vm

    def _windows_is_vm(self) -> bool:
        """Checking if Windows system is Virtual Machine"""
        return any(
            [
                self._check_windows_serial(),
                self._check_windows_platform(env_const.HARDWARE_MANUFACTURER),
                self._check_windows_platform(env_const.HARDWARE_MODEL),
            ]
        )

    def _linux_is_vm(self) -> bool:
        """Checking if Linux system is Virtual Machine"""
        return any(
            [
                self._detect_linux_virt(),
                self._check_linux_platform(env_const.LINUX_HOSTNAMECTL),
                self._check_linux_platform(env_const.LINUX_DMESG),
            ]
        )

    def is_vm(self) -> bool:
        """Checking if supported system is a virtual machine.

        Raises:
            OSError: If system is not supported.

        Returns:
            bool: This is VM.
        """

        platforms: Dict[str, Callable[[], bool]] = {
            "Windows": self._windows_is_vm,
            "Linux": self._linux_is_vm,
        }
        method_call = platforms.get(self.system, None)
        if method_call is None:
            msg = f"Current operating system is not supported: {self.system}."
            raise OSError(msg)

        return method_call()

    def is_pc(self) -> bool:
        """Checking if the system is a physical machine.
        If system is not supported raise OSError.
        """
        return not self.is_vm()

    @staticmethod
    def is_windows() -> bool:
        """Checking if the platform is a Windows.

        Returns:
            bool: The system  is Windows.
        """
        return platform.uname().system == env_const.WINDOWS

    @staticmethod
    def is_linux() -> bool:
        """Checking if the platform is a Linux.

        Returns:
            bool: The system is Linux.
        """
        return platform.uname().system == env_const.LINUX
