from dependency_injector.wiring import Provide, inject
from fastapi import Depends, Request
from fastapi_utils.inferring_router import InferringRouter

from container import Container
from excepts import AuthenticationError
from models.auth import AuthResponse, AuthUserModel, TokenRefresh
from models.base import HTTP_401_RESPONSE_MODEL, HTTP_500_RESPONSE_MODEL
from utils.secure import JWTAuthorization

auth_router = InferringRouter()


@auth_router.post(
    "/auth",
    response_model=AuthResponse,
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def login(
    request: Request,
    user: AuthUserModel,
    service=Depends(Provide[Container.service]),
    auth: JWTAuthorization = Depends(Provide[Container.jwt_auth]),
):
    """
    Method to handle connection request to RDS.

    - **login**: user login.
    - **domain**: userr domain.
    - **password**: user password.
    """

    try:
        is_connected = service.test_connect(user)

        if not is_connected:
            raise AuthenticationError("Учетные данные не прошли проверку")

        user_login, user_domain = user.login, user.domain
        ip_addr = request.client.host
        payload = {"login": user_login, "domain": user_domain, "ip_addr": ip_addr}
        response = auth.create_tokens(payload)

    except Exception as ex:
        raise AuthenticationError(ex) from ex

    return response


@auth_router.post(
    "/auth/refresh",
    response_model=AuthResponse,
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def refresh(params: TokenRefresh, auth: JWTAuthorization = Depends(Provide[Container.jwt_auth])):
    """
    Method for handling a request to refresh JWT-token.

    - **refresh_token**: refresh token for update access token.
    """
    refresh_token = params.refresh_token

    return auth.refresh(refresh_token)
