import logging
from threading import Thread

from fastapi import BackgroundTasks, Depends, status
from fastapi.responses import JSONResponse
from fastapi_utils.cbv import cbv
from fastapi_utils.inferring_router import InferringRouter

import _version
from core.secure import check_health_auth
from env_config import SSA_URL_DELEGATE, SSA_URL_DELEGATE_NAME
from models.base import (
    HTTP_401_RESPONSE_MODEL_HEALTH,
    BaseDelegateModel,
    HealthCheckResponse,
    VersionCheckResponse,
)
from services.agent import AgentService
from services.healthcheck import HealthCheck
from utils.health_check import check_health_inject

logger = logging.getLogger(__name__)

# Const for static app status response
RESPONSE_OK = {"status": "ok"}

PATH_DELEGATE = "/agent/delegate"
base_router = InferringRouter()


@base_router.get(
    "/health",
    response_model=HealthCheckResponse,
    responses={**HTTP_401_RESPONSE_MODEL_HEALTH},
)
def healthcheck(
    service: HealthCheck = Depends(
        check_health_inject,
    ),
    auth=Depends(check_health_auth),
):
    """
    Handler to check the healthcheck of the server.
    """
    if not auth:
        return JSONResponse(
            status_code=status.HTTP_401_UNAUTHORIZED,
            content={"description": "Authorization information is missing or invalid"},
        )
    answer = {
        "status": "pass",
        "version": f"{_version.__version__}",
        "output": "",
        "description": "termidesk-session-agent",
    }
    try:
        status_service, error_service = service.status_service()
        if not status_service:
            answer["status"] = "fail"
            answer["output"] = f"{error_service} completed with an error"
    except Exception as error:
        answer["status"] = "fail"
        answer["output"] = str(error)
    return answer


@cbv(base_router)
class BaseHandler:
    @base_router.get("/version", response_model=VersionCheckResponse)
    def version_check(self):
        """
        Method for handling a request to get version.
        """
        version = _version.__version__

        return {"version": version}

    @base_router.post(PATH_DELEGATE)
    def delegate(self, params: BaseDelegateModel, background_tasks: BackgroundTasks) -> dict:
        """
        Method for manage agent work delegate.

        - **url**: delegate url.
        """

        request_body = params.json()
        log_body = (
            "VM Agent has been notified of the delegation "
            f"acceptance from SSA. HTTP 200 POST '{PATH_DELEGATE}' "
            f"{SSA_URL_DELEGATE_NAME}:'{SSA_URL_DELEGATE}' "
            f"request body:'{request_body}'"
        )

        # accept work delegate from agent
        process = AgentService(params.url).process
        background_process = Thread(target=process)
        background_process.start()

        background_tasks.add_task(logger.info, log_body)

        return JSONResponse(status_code=status.HTTP_200_OK, content=RESPONSE_OK)
