import logging

from dependency_injector.wiring import Provide, inject
from fastapi import Depends, Header
from fastapi_utils.inferring_router import InferringRouter

from container import Container
from core.secure import check_auth
from excepts import AuthenticationError, OSAPIError
from models.auth import OTPTokenResponse
from models.base import HTTP_401_RESPONSE_MODEL, HTTP_500_RESPONSE_MODEL, HTTP_501_RESPONSE_MODEL
from models.meta import (
    MetaApps,
    MetaAvailableApps,
    MetaDeleteApp,
    MetaInstallRdsRole,
    MetaLogin,
    MetaLogout,
    MetaLogoutModel,
    MetaProgramRequestModel,
    MetaPublishApp,
    MetaPublishAppModel,
    MetaSessionModel,
    MetaSessionRequest,
    MetaSessions,
    MetaTest,
    MetaTestConnect,
    MetaUserModel,
)
from services.meta import MetaService
from utils.base import logged_gusid, logged_ursi
from utils.secure import OTPAuthorization, decrypt_password

logger = logging.getLogger(__name__)
meta_router = InferringRouter()

"""
Class for handling requests for Meta

Handled errors:
- OSAPIError (из модуля excepts): the error that occurs
  when executing a request API ОС
- AuthenticationError (из модуля excepts): the error that occurs
  during authentication
"""


@meta_router.get("/test", response_model=MetaTest)
@inject
def check_instance(service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Checking the health of the meta server
    """
    service_name = service.name

    return {"service_name": service_name}


@meta_router.get(
    "/sessions",
    dependencies=[Depends(check_auth)],
    response_model=MetaSessions,
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def sessions(service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Getting sessions list
    """
    try:
        user_sessions = service.get_sessions()
        response = {"sessions": user_sessions}
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return response


@meta_router.post(
    "/login",
    response_model=MetaLogin,
    responses={
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def login(params: MetaUserModel, service: MetaService = Depends(Provide[Container.meta_service])):
    """
    User authorization

    - **login**: user login
    - **domain**: user domain
    - **password**: user password
    """
    try:
        is_authenticated = service.login_user(params)

        response = {"auth_done": is_authenticated}
    except Exception as ex:
        raise AuthenticationError(ex) from ex

    return response


@meta_router.post(
    "/testconnect",
    response_model=MetaTestConnect,
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def test_connect(params: MetaUserModel, service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Checking connect user

    - **login**: user login
    - **domain**: user domain
    - **password**: user password
    """
    try:
        is_connected = service.test_connect(params)
        response = {"is_connected": is_connected}
    except Exception as ex:
        raise AuthenticationError(ex) from ex

    return response


@meta_router.post(
    "/logout",
    dependencies=[Depends(check_auth)],
    response_model=MetaLogout,
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@logged_ursi(logger)
@inject
def logout_user(
    params: MetaLogoutModel,
    service: MetaService = Depends(Provide[Container.meta_service]),
    X_REQUEST_ID: str = Header(include_in_schema=False),
    X_TDSK_URSI: str = Header(default=None),
):
    """
    Disconnecting user

    - **session_id**: session id for logout
    - **X-REQUEST-ID**: not required request header. (generated middleware)
    - **X-TDSK-URSI**: header is expected, but may not be available yet.
    """
    try:
        answer = service.logout_user(params)

        response = {"disconnect_done": answer}
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return response


@meta_router.get(
    "/published_apps",
    dependencies=[Depends(check_auth)],
    response_model=MetaApps,
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def published_apps(service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Getting published applications
    """
    try:
        apps = service.get_remote_apps()

        response = {"apps": apps}
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return response


@meta_router.post(
    "/applications",
    response_model=MetaPublishApp,
    dependencies=[Depends(check_auth)],
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_501_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def publish_app(
    params: MetaPublishAppModel, service: MetaService = Depends(Provide[Container.meta_service])
):
    """
    Adding an application on Windows Server to the list of available

    - **name**: application name
    - **path**: path to application
    """
    try:
        service.publish_app(params)

        response = {"is_created": True}
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return response


@meta_router.delete(
    "/applications/{name}",
    response_model=MetaDeleteApp,
    dependencies=[Depends(check_auth)],
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_501_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def delete_app(name: str, service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Removing an application on Windows Server from the list of available

    - **name**: application name
    """
    try:
        service.delete_app(name)

        response = {"is_deleted": True}
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return response


@meta_router.get(
    "/available_apps",
    dependencies=[Depends(check_auth)],
    response_model=MetaAvailableApps,
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_501_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def check_available_apps(service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Getting publishable Windows Server applications
    """
    try:
        app_list = service.get_list_startapps()
        response = {"available_apps": app_list}
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return response


@meta_router.post(
    "/sessionrequest",
    response_model=MetaSessionRequest,
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_501_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@logged_gusid(logger)
@logged_ursi(logger)
@inject
def session_request(
    data_session: MetaSessionModel,
    otp_auth: OTPAuthorization = Depends(Provide[Container.otp_auth]),
    service: MetaService = Depends(Provide[Container.meta_service]),
    X_REQUEST_ID: str = Header(include_in_schema=False),
    X_TDSK_GUSID: str = Header(default=None),
    X_TDSK_URSI: str = Header(default=None),
):
    """
    Method for creating user session in meta STAL.

    - **username**: user login.
    - **domain**: user domain.
    - **password**: user password.
    - **data**: options for creating a session.
    - **server_id**: pre-shared key.
    - **X-REQUEST-ID**: not required request header. (generated middleware)
    - **X-TDSK-GUSID**: header is expected, but may not be available yet.
    - **X-TDSK-URSI**: header is expected, but may not be available yet.
    """
    token = data_session.data.get("token")

    otp_auth.verify(token)
    password = decrypt_password(password_encrypted=data_session.password, server_id=data_session.server_id)
    params = MetaUserModel(
        login=data_session.username,
        domain=data_session.domain,
        password=password,
    )
    try:
        result = service.session_request(params)
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return {"session_id": result}


@meta_router.post(
    "/programrequest",
    response_model=MetaSessionRequest,
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_501_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@logged_gusid(logger)
@logged_ursi(logger)
@inject
def program_request(
    data_session: MetaSessionModel,
    otp_auth: OTPAuthorization = Depends(Provide[Container.otp_auth]),
    service: MetaService = Depends(Provide[Container.meta_service]),
    X_REQUEST_ID: str = Header(include_in_schema=False),
    X_TDSK_GUSID: str = Header(default=None),
    X_TDSK_URSI: str = Header(default=None),
):
    """
    Method for creating user session in meta STAL in single program mode.

    - **username**: user login.
    - **domain**: user domain.
    - **password**: user password.
    - **data**: options for creating a session.
    - **server_id**: pre-shared key.
    - **X-REQUEST-ID**: not required request header. (generated middleware)
    - **X-TDSK-GUSID**: header is expected, but may not be available yet.
    - **X-TDSK-URSI**: header is expected, but may not be available yet.
    """
    token = data_session.data.get("token")
    program = data_session.data.get("program")

    otp_auth.verify(token)
    password = decrypt_password(password_encrypted=data_session.password, server_id=data_session.server_id)
    params = MetaProgramRequestModel(
        login=data_session.username,
        domain=data_session.domain,
        password=password,
        program=program,
    )
    try:
        result = service.program_request(params)
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return {"session_id": result}


@meta_router.post(
    "/install_rds_role",
    dependencies=[Depends(check_auth)],
    response_model=MetaInstallRdsRole,
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_501_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def install_rds_role(service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Method for handling a request to install rds role on powershell
    """

    return service.install_rds_role()


@meta_router.get(
    "/otp",
    response_model=OTPTokenResponse,
    dependencies=[Depends(check_auth)],
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def otp_token(auth: OTPAuthorization = Depends(Provide[Container.otp_auth])):
    """
    Method to handle a request get OTP token.
    """
    token = auth.create_tokens()
    return {"token": token}
