from dependency_injector.wiring import Provide, inject
from fastapi import Depends
from fastapi_utils.inferring_router import InferringRouter
from pywintypes import error as pywin_error
from wmi import x_wmi

from container import Container
from core.secure import check_auth
from excepts import AuthenticationError, OSAPIError
from models.base import HTTP_401_RESPONSE_MODEL, HTTP_500_RESPONSE_MODEL
from models.rds import (
    CheckInstanceResponse,
    CollectionInfoResponse,
    CollectionsResponse,
    LoginResponse,
    LogoutUserResponse,
    RdsLogoutModel,
    RdsRemoteAppModel,
    RdsRemoteAppResponse,
    RdsUserModel,
    RemoteAppUsersModel,
    RemoteAppUsersResponse,
    SessionsResponse,
    TestConnectResponse,
)
from services.rds import RdsService
from utils.rds import check_instance_is_rds

rds_router = InferringRouter()


@rds_router.get(
    "/rds/test",
    response_model=CheckInstanceResponse,
    responses={
        **HTTP_500_RESPONSE_MODEL,
    },
)
def check_rds_instance():
    """
    Method for handling server rds check request.
    """
    try:
        result = check_instance_is_rds()
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return {"is_rds": result}


@rds_router.get(
    "/rds/sessions",
    response_model=SessionsResponse,
    dependencies=[Depends(check_auth)],
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def rds_sessions(service: RdsService = Depends(Provide[Container.service])):
    """
    Method for handling a request to get rds user sessions.
    """
    try:
        result = service.get_sessions()
    except x_wmi as ex:
        raise OSAPIError(ex) from ex

    return {"sessions": result}


@rds_router.post(
    "/rds/login",
    response_model=LoginResponse,
    responses={
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def rds_login(params: RdsUserModel, service: RdsService = Depends(Provide[Container.service])):
    """
    User authorization.

    - **login**: user login.
    - **domain**: user domain.
    - **password**: user password.
    """
    try:
        result = service.login_user(params)
    except pywin_error as ex:
        raise AuthenticationError(ex) from ex

    return {"auth_done": result}


@rds_router.post(
    "/rds/testconnect",
    response_model=TestConnectResponse,
    responses={
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def rds_test_connect(params: RdsUserModel, service: RdsService = Depends(Provide[Container.service])):
    """
    Checking connect user.

    - **login**: user login.
    - **domain**: user domain.
    - **password**: user password.
    """
    try:
        result = service.test_connect(params)
    except pywin_error as ex:
        raise AuthenticationError(ex) from ex

    return {"is_connected": result}


@rds_router.post(
    "/rds/disconnect",
    response_model=LogoutUserResponse,
    dependencies=[Depends(check_auth)],
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def rds_logout_user(params: RdsLogoutModel, service: RdsService = Depends(Provide[Container.service])):
    """
    Disconnecting user.

    - **session_id**: session id for logout.
    """
    try:
        result = service.logout_user(params)
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return {"disconnect_done": result}


@rds_router.get(
    "/rds/collections",
    response_model=CollectionsResponse,
    dependencies=[Depends(check_auth)],
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def rds_collections(service: RdsService = Depends(Provide[Container.service])):
    """
    Method for handling a request to get rds collections.
    """
    try:
        result = service.get_collections()
    except x_wmi as ex:
        raise OSAPIError(ex) from ex

    return {"collections": result}


@rds_router.post(
    "/rds/apps",
    response_model=RdsRemoteAppResponse,
    dependencies=[Depends(check_auth)],
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def rds_apps(params: RdsRemoteAppModel, service: RdsService = Depends(Provide[Container.service])):
    """
    Method for handling a request to get published remote apps.

    - **collection**: collection name.
    """
    try:
        result = service.get_remote_apps(params)
    except x_wmi as ex:
        raise OSAPIError(ex) from ex

    return {"apps": result}


@rds_router.post(
    "/rds/apps/users",
    response_model=RemoteAppUsersResponse,
    dependencies=[Depends(check_auth)],
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def rds_app_users(params: RemoteAppUsersModel, service: RdsService = Depends(Provide[Container.service])):
    """
    Method for handling a request to getting users that have access
    to published app by collection name and remote app alias.

    - **collection**: collection name.
    - **app_alias**: application alias.
    """
    try:
        result = service.get_remote_app_users(params)
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return {"app_users": result}


@rds_router.get(
    "/rds/collection_info/{collection_name}",
    response_model=CollectionInfoResponse,
    dependencies=[Depends(check_auth)],
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def collection_info(collection_name: str, service: RdsService = Depends(Provide[Container.service])):
    """
    Method for handling a request to get collection info by collection name.

    - **collection_name**: name collection apps.
    """
    try:
        result = service.get_collection_info(collection_name)
    except x_wmi as ex:
        raise OSAPIError(ex) from ex

    return {"collection_info": result}
