import logging

import dbus
from dependency_injector.wiring import Provide, inject
from fastapi import Depends, Header
from fastapi_utils.inferring_router import InferringRouter

from container import Container
from core.secure import check_auth
from excepts import AuthenticationError, OSAPIError
from models.auth import OTPTokenResponse
from models.base import HTTP_401_RESPONSE_MODEL, HTTP_500_RESPONSE_MODEL
from models.stal import (
    AppsResponse,
    DisconnectUserResponse,
    LoginUserResponse,
    ProgramRequestResponse,
    SessionRequestResponse,
    SessionsResponse,
    ShutdownUserResponse,
    StalLogoutModel,
    StalRemoteAppModel,
    StalSessionModel,
    StalUserModel,
    StatusResponse,
    TestConnectResponse,
    VersionResponse,
)
from services.stal import StalService
from utils.base import logged_gusid, logged_ursi
from utils.secure import OTPAuthorization, decrypt_password

logger = logging.getLogger(__name__)
stal_router = InferringRouter()


@stal_router.get("/stal/versions", response_model=VersionResponse)
@inject
def version(service: StalService = Depends(Provide[Container.service])):
    """
    Method for gets STAL version.
    """
    try:
        result = service.version()
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return {"version": result, "message": "OK!"}


@stal_router.get(
    "/rds/test",  # FIXME: "/stal/test"
    response_model=StatusResponse,
    responses={
        **HTTP_500_RESPONSE_MODEL,
    },
)
@inject
def status(service: StalService = Depends(Provide[Container.service])):
    """
    Method for gets STAL status.
    """
    try:
        result = service.status()
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return {"is_stal": result}


@stal_router.get(
    "/rds/sessions",  # FIXME: "/stal/sessions"
    response_model=SessionsResponse,
    dependencies=[Depends(check_auth)],
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def sessions(service: StalService = Depends(Provide[Container.service])):
    """
    Method for gets a list of running sessions.
    """
    try:
        result = service.get_sessions()
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return {"sessions": result}


@stal_router.post(
    "/rds/login",  # FIXME: "/stal/login"
    response_model=LoginUserResponse,
    responses={
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def login_user(user: StalUserModel, service: StalService = Depends(Provide[Container.service])):
    """
    Method for authentication user.

    - **login**: user login.
    - **domain**: user domain.
    - **password**: user password.
    """
    try:
        result = service.login_user(user)
    except Exception as ex:
        raise AuthenticationError(ex) from ex

    return {"auth_done": result}


@stal_router.post(
    "/stal/disconnect_",
    response_model=DisconnectUserResponse,
    dependencies=[Depends(check_auth)],
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def disconnect_user(user: StalLogoutModel, service: StalService = Depends(Provide[Container.service])):
    """
    Method for disconnect user.

    - **session_id**: session id for disconnect.
    """
    try:
        result = service.disconnect_user(user)
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return {"message": "OK!" if result else "BAD!"}


@stal_router.post(
    "/rds/disconnect",  # FIXME: "/stal/shutdown"
    response_model=ShutdownUserResponse,
    dependencies=[Depends(check_auth)],
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@logged_ursi(logger)
@inject
def shutdown_user(
    user: StalLogoutModel,
    service: StalService = Depends(Provide[Container.service]),
    X_REQUEST_ID: str = Header(include_in_schema=False),
    X_TDSK_URSI: str = Header(default=None),
):
    """
    Method for shutdown user session.

    - **session_id**: session id for logout.
    - **X-REQUEST-ID**: not required request header. (generated middleware)
    - **X-TDSK-URSI**: header is expected, but may not be available yet.
    """
    try:
        result = service.logout_user(user)
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return {"disconnect_done": result}


@stal_router.post(
    "/rds/testconnect",  # FIXME: "/stal/testconnect"
    response_model=TestConnectResponse,
    responses={
        **HTTP_500_RESPONSE_MODEL,
    },
)
@inject
def test_connect(user: StalUserModel, service: StalService = Depends(Provide[Container.service])):
    """
    Method for check user in STAL.

    - **login**: user login.
    - **domain**: user domain.
    - **password**: user password.
    """
    try:
        result = service.test_connect(user)
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return {"is_connected": result}


@stal_router.post(
    "/stal/sessionrequest",
    response_model=SessionRequestResponse,
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@logged_gusid(logger)
@logged_ursi(logger)
@inject
def session_request(
    data_session: StalSessionModel,
    otp_auth: OTPAuthorization = Depends(Provide[Container.otp_auth]),
    service: StalService = Depends(Provide[Container.service]),
    X_REQUEST_ID: str = Header(include_in_schema=False),
    X_TDSK_GUSID: str = Header(default=None),
    X_TDSK_URSI: str = Header(default=None),
):
    """
    Method for creating user session in STAL.

    - **username**: user login.
    - **domain**: user domain.
    - **password**: user password.
    - **data**: options for creating a session.
    - **server_id**: pre-shared key.
    - **X-REQUEST-ID**: not required request header. (generated middleware)
    - **X-TDSK-GUSID**: header is expected, but may not be available yet.
    - **X-TDSK-URSI**: header is expected, but may not be available yet.
    """
    token = data_session.data.get("token")

    otp_auth.verify(token)
    password = decrypt_password(password_encrypted=data_session.password, server_id=data_session.server_id)
    params = StalUserModel(
        login=data_session.username,
        domain=data_session.domain,
        password=password,
    )
    try:
        result = service.session_request(params)
    except dbus.exceptions.DBusException as ex:
        raise OSAPIError(ex) from ex

    return {"session_id": result}


@stal_router.post(
    "/stal/programrequest",
    response_model=ProgramRequestResponse,
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@logged_gusid(logger)
@logged_ursi(logger)
@inject
def program_request(
    data_session: StalSessionModel,
    otp_auth: OTPAuthorization = Depends(Provide[Container.otp_auth]),
    service: StalService = Depends(Provide[Container.service]),
    X_REQUEST_ID: str = Header(include_in_schema=False),
    X_TDSK_GUSID: str = Header(default=None),
    X_TDSK_URSI: str = Header(default=None),
):
    """
    Method for creating user session in STAL in single program mode.

    - **username**: user login.
    - **domain**: user domain.
    - **password**: user password.
    - **data**: options for creating a session.
    - **server_id**: pre-shared key.
    - **X-REQUEST-ID**: not required request header. (generated middleware)
    - **X-TDSK-GUSID**: header is expected, but may not be available yet.
    - **X-TDSK-URSI**: header is expected, but may not be available yet.
    """
    token = data_session.data.get("token")
    program = data_session.data.get("program")

    otp_auth.verify(token)
    password = decrypt_password(password_encrypted=data_session.password, server_id=data_session.server_id)
    params = StalRemoteAppModel(
        login=data_session.username,
        domain=data_session.domain,
        password=password,
        program=program,
    )
    try:
        result = service.program_request(params)
    except dbus.exceptions.DBusException as ex:
        raise OSAPIError(ex) from ex

    return {"session_id": result}


@stal_router.get(
    "/stal/apps",
    response_model=AppsResponse,
    dependencies=[Depends(check_auth)],
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def stal_apps(service: StalService = Depends(Provide[Container.service])):
    """
    Method for to get published remote apps.
    """
    try:
        result = service.get_remote_apps()
    except Exception as ex:
        raise OSAPIError(ex) from ex

    return {"apps": result}


@stal_router.get(
    "/stal/otp",
    response_model=OTPTokenResponse,
    dependencies=[Depends(check_auth)],
    responses={
        **HTTP_500_RESPONSE_MODEL,
        **HTTP_401_RESPONSE_MODEL,
    },
)
@inject
def otp_token(auth: OTPAuthorization = Depends(Provide[Container.otp_auth])):
    """
    Method to handle a request get OTP token.
    """
    token = auth.create_tokens()
    return {"token": token}
