import json
from typing import Any, Dict, List

import dbus

from constants.common_const import ExceptionMessages
from constants.linux.constants import ConstStal
from excepts import OSAPIError
from models.meta import MetaLogoutModel, MetaProgramRequestModel, MetaPublishAppModel, MetaUserModel
from utils.base import extract_ip


class MetaService:
    """
    Сервисный слой для взаимодействия со STAL
    """

    __bus = None
    __stal = None
    __stal_iface = None
    __session_info: dict = {}
    __name = "stal"

    # Class implemented via Singleton
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super().__new__(cls)
        return cls._instance

    # DI для проброса констант
    def __init__(self, const: ConstStal):
        self.DEPTH = const.DEPTH
        self.DISPLAY_WIDTH = const.DISPLAY_WIDTH
        self.DISPLAY_HEIGHT = const.DISPLAY_HEIGHT
        self.SESSION_STATES = const.SESSION_STATES
        self.TIMEOUT = const.TIMEOUT

    @property
    def stal_iface(self) -> dbus.Interface:
        if not self.__stal_iface:
            self.__bus = dbus.SystemBus()
            self.__stal = self.__bus.get_object("ru.uveon.stal", "/ru/uveon/stal")
            self.__stal_iface = dbus.Interface(self.__stal, "ru.uveon.stal")
            return self.__stal_iface

        return self.__stal_iface

    @property
    def name(self) -> str:
        return self.__name

    def get_session_info(self, session_id: dbus.UInt32) -> dbus.String:
        """Gets session information.

        Args:
            session_id: Session id for more information.

        Returns:
            dbus.String: Json session information.
        """
        session_info: dbus.String = self.stal_iface.stalSessionInfoJson(session_id)
        return session_info

    def __load_session_info(self, session_id: dbus.UInt32) -> None:
        """Records session information.

        Args:
            session_id: Session id for more information.
        """
        info: dbus.String = self.get_session_info(session_id)
        self.__session_info = json.loads(info)

    def __clear_session_info(self) -> None:
        """Removes session information."""
        self.__session_info = {}

    def __get_session_state(self) -> str:
        """Get session state.

        A - Active.
        I - Inactive.

        Returns:
            str: Session state A OR I.
        """
        is_session_active = self.__session_info.get("remote:connected")
        if is_session_active is None:
            # FIXME:Для stal меньше v.1.1 проверяем наличие ключа 'pid3'
            is_session_active = self.__session_info["pid3"] != 0
        return self.SESSION_STATES[is_session_active]

    def __get_session_is_app(self) -> bool:
        """Get the value of whether the session is an application.

        Parameter value 'seamless:mode':
        if mode == false - terminal.
        if mode == true - app.

        Returns:
            bool: The session is app.
        """
        is_app_mode = self.__session_info["seamless:mode"]
        if isinstance(is_app_mode, str):  # special for STAL V0.3
            is_app_mode = is_app_mode == "true"
        return is_app_mode

    def __check_session_create(self, session_id: dbus.Int32) -> None:
        """Check session STAL creation.

        Args:
            session_id: Session id.

        Raises:
            OSAPIError: STAL session not created!
        """
        if session_id == -1:
            raise OSAPIError(ExceptionMessages.STAL_SESSION_NOT_CREATED)

    def __check_run_transport(self, result: dbus.Boolean) -> None:
        """Check run RDP transport.

        Args:
            result: result of RDP transport.

        Raises:
            OSAPIError: Transport RDP did not start!
        """
        if not result:
            raise OSAPIError(ExceptionMessages.TRANSPORT_RDP_DID_NOT_START)

    def get_sessions(self) -> List[Dict[str, Any]]:
        """Method for gets STAL sessions.

        Returns:
            List[Dict[str, Any]]: Sessions data.
        """
        # Example:
        # dbus.Array(
        #     [
        #         dbus.Struct(
        #             (dbus.String("test1"), dbus.UInt32(59101)), signature=None
        #         )
        #     ],
        #     signature=dbus.Signature("(su)"),
        # )
        dbus_sessions: dbus.Array[dbus.Struct] = self.stal_iface.stalSessions()
        host_server = extract_ip()
        sessions = []
        for session in dbus_sessions:
            username, session_id = session
            self.__load_session_info(session_id)
            state = self.__get_session_state()
            is_desktop = not self.__get_session_is_app()
            self.__clear_session_info()

            sessions.append(
                {
                    "username": username,
                    "session_id": session_id,
                    "host_server": host_server,  # for logoff in future
                    "state": state,
                    "is_desktop": is_desktop,
                }
            )

        return sessions

    def test_connect(self, params: MetaUserModel) -> dbus.Boolean:
        """Method for authenticating user credentials.

        Args:
            params: Options for login user a session.

        Returns:
            dbus.Boolean: The user's credentials are authentic.
        """
        is_success: dbus.Boolean = self.stal_iface.stalTestConnect(
            params.login,  # user
            params.domain,  # domain
            params.password,  # pass
        )
        return is_success

    def login_user(self, params: MetaUserModel) -> dbus.Boolean:
        """Method for start of a new user session.

        Args:
            params: Options for login user a session.

        Returns:
            dbus.Boolean: User session login.
        """
        is_success: dbus.Boolean = self.stal_iface.stalLogin(
            params.login,  # user
            params.domain,  # domain
            params.password,  # pass
        )
        return is_success

    def logout_user(self, params: MetaLogoutModel) -> dbus.Boolean:
        """Method for shutdown user session.

        Session closes assync ~600ms.

        Args:
            params: Options for logout user a session.

        Returns:
            dbus.Boolean: User session logout.
        """
        is_success: dbus.Boolean = self.stal_iface.stalSessionShutdown(
            params.session_id,
        )
        return is_success

    def session_request(self, params: MetaUserModel) -> dbus.Int32:
        """Method for creating user session in STAL.

        Args:
            params: Options for starting a session.

        Returns:
            dbus.Int32: Session id.
        """
        # session_id может быть -1, если сессия не была создана
        data = json.dumps(
            {
                "username": params.login,
                "domain": params.domain,
                "password": params.password,
                "depth": self.DEPTH,
                "width": self.DISPLAY_WIDTH,
                "height": self.DISPLAY_HEIGHT,
                "transport:run": False,
            }
        )
        session_id: dbus.Int32 = self.stal_iface.stalSessionRequest(data)
        self.__check_session_create(session_id)
        result: dbus.Boolean = self.stal_iface.stalConnectRdpTransport(
            session_id,  # session
            self.TIMEOUT,  # timeout
        )
        self.__check_run_transport(result)
        return session_id

    def program_request(self, params: MetaProgramRequestModel) -> dbus.Int32:
        """Method for creating user session in STAL in single program mode.

        Args:
            params: Options for starting a session.

        Returns:
            dbus.Int32: Session id.
        """
        data = json.dumps(
            {
                "username": params.login,
                "domain": params.domain,
                "password": params.password,
                "program": params.program,
                "depth": self.DEPTH,
                "width": self.DISPLAY_WIDTH,
                "height": self.DISPLAY_HEIGHT,
                "transport:run": False,
            }
        )
        session_id: dbus.Int32 = self.stal_iface.stalSessionRequest(data)
        self.__check_session_create(session_id)
        result: dbus.Boolean = self.stal_iface.stalConnectRdpTransport(
            session_id,  # session
            self.TIMEOUT,  # timeout
        )
        self.__check_run_transport(result)
        return session_id

    def get_remote_apps(self) -> dbus.Array:
        """Getting published remote apps from STAL.

        Returns:
            dbus.Array: Array of programs.
        """
        remote_apps: dbus.Array = self.stal_iface.stalProgramList()
        return remote_apps

    def publish_app(self, params: MetaPublishAppModel) -> None:
        """Application Publishing method on STAL.

        Args:
            params: publish params.

        Raises:
            NotImplementedError: The method is not intended to be called on this OS.
        """
        raise NotImplementedError("The method is not intended to be called on this OS.")

    def delete_app(self, name: str) -> None:
        """Deleting an application to STAL.

        Args:
            name: application name.

        Raises:
            NotImplementedError: The method is not intended to be called on this OS.
        """
        raise NotImplementedError("The method is not intended to be called on this OS.")

    def get_list_startapps(self) -> None:
        """Getting list of apps from Start Menu.

        Available apps for publish.

        Raises:
            NotImplementedError: The method is not intended to be called on this OS.
        """
        raise NotImplementedError("The method is not intended to be called on this OS.")

    def install_rds_role(self) -> None:
        """Configuring the RDS Role.

        Raises:
            NotImplementedError: The method is not intended to be called on this OS.
        """
        raise NotImplementedError("The method is not intended to be called on this OS.")
