﻿import os
from typing import Dict, List, Optional, Union

import win32api
import win32security

import eventlog
from constants.windows.constants import ConstRds
from models.rds import RdsLogoutModel, RdsRemoteAppModel, RdsUserModel, RemoteAppUsersModel
from utils.rds import WMIHandler, psh_request_handler


class RdsService:
    """
    Сервис взаимодействия с RDS.

    Handled errors:
    - OSAPIError (из модуля excepts): the error that occurs
      when executing a request API ОС.
    - AuthenticationError (из модуля excepts): the error that occurs
      during authentication.
    """

    def __init__(self, wmi_handler: WMIHandler, const: ConstRds):
        self._wmi_handler = wmi_handler
        self.GET_REMOTE_APP_USERS_CMD = const.GET_REMOTE_APP_USERS_CMD
        self.LOGOFF_USER_CMD = const.LOGOFF_USER_CMD

    def get_sessions(self) -> List[Dict[str, Union[str, int, bool]]]:
        """Method returning active rds sessions.

        Returns:
            List[Dict[str, Union[str, int, bool]]]: Sessions data.
        """
        return self._wmi_handler.get_sessions()

    def login_user(self, params: RdsUserModel) -> bool:
        """Authorization function in rds.

        Args:
            params: Connect params.

        Returns:
            bool: Successful user login.
        """
        # LogonUser() return a handle to a token that represents the specified user.
        # Returned handle is a primary token that you can use in calls to the
        # CreateProcessAsUser() function.
        # If you specify the LOGON32_LOGON_NETWORK flag, LogonUser() returns
        # an impersonation token that you cannot use in CreateProcessAsUser()
        # unless you call DuplicateTokenEx() to convert it to a primary token.
        # LOGON32_PROVIDER_WINNT40: Use the NTLM logon provider.
        token = win32security.LogonUser(
            params.login,
            params.domain,
            params.password,
            win32security.LOGON32_LOGON_NETWORK,
            win32security.LOGON32_PROVIDER_WINNT40,
        )
        is_authenticated = bool(token)
        win32api.CloseHandle(token)
        eventlog.write_event("Auth is done.", "INFO")
        return is_authenticated

    def test_connect(self, params: RdsUserModel) -> bool:
        """Method that return response of trying to connect to rds.

        Args:
            params: Options for login user a session.

        Returns:
            bool: The user's credentials are authentic.
        """
        # TODO Пока не понятно как правильно делать проверку тестового подключения по RDP.
        # Оставим текущую логику закомментированной.

        # RDS_HOSTNAME = '192.168.3.13'
        # login = params.login
        # pwd = params.password

        # os.system(
        #     'cmd/c "cmdkey /generic:"{hostname}" /user:"{login}" \
        #         /pass:"{pwd}""'.format(hostname=RDS_HOSTNAME, login=login, pwd=pwd)
        #     )
        # os.system('cmd/c "mstsc /v:{hostname}"'.format(hostname=RDS_HOSTNAME))
        # os.system('cmd/c "cmdkey /delete:{hostname}"'.format(hostname=RDS_HOSTNAME))

        # Временное решение делать проверку посредством попытки авторизации на AD.
        return self.login_user(params)

    def status(self) -> bool:
        """
        At the moment, verification is not necessary

        Returns:
            bool: True
        """
        return True

    @staticmethod
    def __get_vhost_by_session_id(session_id: int, sessions: List[dict]) -> Optional[str]:
        """
        Method for return vhost by session id.

        Args:
            session_id: Session id.
            sessions: List with sessions.

        Returns:
            Optional[str]: Vhost.
        """
        for session in sessions:
            if session["session_id"] == session_id:
                return session["host_server"]
        return None

    def logout_user(self, params: RdsLogoutModel) -> bool:
        """
        Method for logout users in RDS.

        Args:
            params: object of RdsLogoutModel.

        Returns:
            bool: Answer about logout user, in bool format.
        """
        sessions = self._wmi_handler.get_sessions()
        session_id = params.session_id

        host_server = self.__get_vhost_by_session_id(session_id, sessions)

        if host_server:
            os.system(f'cmd/c "echo y|LOGOFF /SERVER:{host_server} /VM {session_id}"')
            eventlog.write_event(f"Logout user {session_id}", "INFO")
            return True

        eventlog.write_event("User is not disconnected. Session ID not found", "ERROR")
        return False

    def get_collections(self) -> List[Dict[str, str]]:
        """Getting RDS collections.

        Returns:
            List[Dict[str, str]]: Information about collections.
        """
        return self._wmi_handler.get_collections()

    def get_remote_app_users(self, params: RemoteAppUsersModel) -> List[str]:
        """Get remote app user.

        Getting users that have access to published app by collection name and remote app alias.

        Args:
            params: User information.

        Returns:
            List[str]: List of applications available to the user.
        """
        collection = params.collection
        app_alias = params.app_alias
        # Getting users list using powershell handler. Todo: soon to be replaced with WMI handler.
        output = psh_request_handler(
            self.GET_REMOTE_APP_USERS_CMD.format(collection, app_alias)
        )  # FIXME[aap]: В текущей реализации невозможно понять, что команда завершилась с ошибкой
        return [users.lstrip("UserGroups : ") for users in output]

    def get_remote_apps(self, params: RdsRemoteAppModel) -> List[Dict[str, str]]:
        """Getting published remote apps from RDS by collection name.

        Args:
            params: Collection params.

        Returns:
            List[Dict[str, str]]: Information about sessions.
        """
        collection = params.collection
        return self._wmi_handler.get_remote_apps(collection)

    def get_collection_info(self, collection_name: str) -> Dict[str, str]:
        """Getting collection info by collection name.

        Args:
            collection_name: contains collection name.

        Returns:
            dict: The gateway_name in dict format.
        """
        return self._wmi_handler.get_collection_info(collection_name)
