import functools
import logging
import socket

from constants.common_const import ConstConfig as const


def extract_ip() -> str:
    """Get machine ip address.

    Returns:
        str: Local IP address.
    """
    st = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        st.connect((const.INTRANET_IP, 1))
        ip, _ = st.getsockname()  # type: (str, int)
    except Exception:
        ip = const.LOCAL_IP
    finally:
        st.close()
    return ip


def add_gusid_to_log(logger: logging.Logger, x_request_id: str, x_tdsk_gusid: str) -> None:
    """
    Logging GUSID.

    Args:
        logger: instance module logger.
        x_request_id: string request identificator.
        x_tdsk_gusid: string GUSID value.
    """

    if x_tdsk_gusid:
        logger.info(f'[{x_request_id}] В заголовках запроса получен GUSID: "{x_tdsk_gusid}"')
    else:
        logger.warning(f'[{x_request_id}] "X-TDSK-GUSID" отсутствует в заголовках запроса')


def logged_gusid(logger: logging.Logger):
    """
    Decorator to logged GUSID.

    Args:
        logger: instance logging in module.

    Returns:
        function: decorated func.
    """

    def inner_deco(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            x_request_id = kwargs.get("X_REQUEST_ID", "undefined")
            x_tdsk_gusid = kwargs.get("X_TDSK_GUSID")
            add_gusid_to_log(logger, x_request_id, x_tdsk_gusid)
            return func(*args, **kwargs)

        return wrapper

    return inner_deco


def add_ursi_to_log(logger: logging.Logger, x_request_id: str, x_tdsk_ursi: str) -> None:
    """
    Logging URSI.

    Args:
        logger: instance module logger.
        x_request_id: string request identificator.
        x_tdsk_ursi: string URSI value.
    """

    if x_tdsk_ursi:
        logger.info(f'[{x_request_id}] В заголовках запроса получен URSI: "{x_tdsk_ursi}"')
    else:
        logger.warning(f'[{x_request_id}] "X-TDSK-URSI" отсутствует в заголовках запроса')


def logged_ursi(logger: logging.Logger):
    """
    Decorator to logged URSI.

    Args:
        logger: instance logging in module.

    Returns:
        function: decorated func.
    """

    def inner_deco(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            x_request_id = kwargs.get("X_REQUEST_ID", "undefined")
            x_tdsk_ursi = kwargs.get("X_TDSK_URSI")
            add_ursi_to_log(logger, x_request_id, x_tdsk_ursi)
            return func(*args, **kwargs)

        return wrapper

    return inner_deco
