import ctypes
import locale
import subprocess
import winreg as wrg
from contextlib import suppress
from itertools import count
from typing import List

from constants.windows.constants import ConstRds
from core.settings import config


class ApplicationRegEdit:
    """Class for application interaction with the registry"""

    def __init__(self, app_name: str = "") -> None:
        self._app_name: str = app_name

    def __create_key(self, name: str, type_, value) -> None:
        """Creates a key in the registry.

        Args:
            name: Key name.
            type_: Value type.
            value: Value.
        """
        root_path = ConstRds.APP_REG_PATH + "\\" + self._app_name

        with wrg.CreateKey(ConstRds.REG_LOCATION, root_path) as reg:
            wrg.SetValueEx(reg, name, ConstRds.RESERVED, type_, value)

    def _create_dword_key(self, name: str, value) -> None:
        """Creates a REG_DWORD key in the registry.

        Args:
            name: Key name.
            value: Value.
        """
        self.__create_key(name, wrg.REG_DWORD, value)

    def _create_sz_key(self, name: str, value) -> None:
        """Creates a REG_SZ key in the registry.

        Args:
            name: Key name.
            value: Value.
        """
        self.__create_key(name, wrg.REG_SZ, value)

    def __is_path_exists(self) -> bool:
        """Checks is a path exists.

        Returns:
            bool: The path exists.
        """
        root_path = ConstRds.APP_REG_PATH + "\\" + self._app_name
        try:
            wrg.OpenKey(ConstRds.REG_LOCATION, root_path)
        except FileNotFoundError:
            return False
        else:
            return True

    def create_app(self, app_path: str) -> None:
        """Creates an entry with application information in the registry.

        Args:
            app_path: Path to the application.

        Raises:
            Exception: An application with the same name already exists.
        """
        if self.__is_path_exists():
            raise Exception("An application with the same name already exists.")

        try:
            self._create_sz_key("Name", self._app_name)
            self._create_sz_key("Path", app_path)
            self._create_sz_key("VPath", app_path)
            self._create_sz_key("RequiredCommandLine", "")
            self._create_dword_key("CommandLineSetting", ConstRds.REG_ENABLED)
            self._create_sz_key("IconPath", app_path)
            self._create_dword_key("IconIndex", ConstRds.REG_DISABLED)
            self._create_dword_key("ShowInTSWA", ConstRds.REG_DISABLED)
        # Todo (ap): пустой except. Необходимо конкретизировать ошибку. После чего удалить "# noqa E722"
        except:  # noqa E722
            with suppress(OSError):
                self.delete_app()
            raise

    def update_app(self, app_path: str) -> None:
        """Updates an application in the registry.

        Args:
            app_path: Path to the application.

        Raises:
            Exception: An application with that name does not exist.
        """
        if not self.__is_path_exists():
            raise Exception("An application with that name does not exist.")

        self._create_sz_key("Path", app_path)
        self._create_sz_key("VPath", app_path)
        self._create_sz_key("IconPath", app_path)

    def delete_app(self) -> None:
        """Removes an application from the registry"""
        with wrg.OpenKey(ConstRds.REG_LOCATION, ConstRds.APP_REG_PATH) as reg:
            wrg.DeleteKey(reg, self._app_name)

    @staticmethod
    def get_apps() -> List[str]:
        """
        Getting published remote apps using regedit

        Returns:
            List[str]: list of published remote apps by their collection name.
        """
        remote_app = []
        with suppress(OSError), wrg.OpenKey(ConstRds.REG_LOCATION, ConstRds.APP_REG_PATH) as access_key:
            remote_app.extend(wrg.EnumKey(access_key, i) for i in count())

        return remote_app


class WindowsConfigurator:
    """
    Configuring the Windows to work with RDS role
    """

    def __init__(self):
        windll = ctypes.windll.kernel32
        # Get current locale OS, e.g. "ru_RU"
        machine_locale = locale.windows_locale[windll.GetUserDefaultUILanguage()]
        # Get language without region, e.g. "ru"
        machine_lang = machine_locale.split("_")[0]

        if machine_lang == "ru":
            self.rds_lang_modifier = ConstRds.MACHINE_LANG["RU"]
        else:
            self.rds_lang_modifier = ConstRds.MACHINE_LANG["EN"]

    @staticmethod
    def powershell_request(command: str) -> subprocess.CompletedProcess:
        """Powershell output handler.

        Args:
            command: Powershell command.

        Returns:
            subprocess.CompletedProcess: The result of executing a command in Powershell.
        """
        return subprocess.run(
            ["powershell", "-Command", command],
            capture_output=True,
            text=True,
            encoding="cp866",
            check=False,
        )

    @staticmethod
    def __create_regedit_key(root_path: str, name: str, type_, value) -> None:
        """Creates a key in the registry.

        Args:
            root_path: Path to directory.
            name: Key name.
            type_: Value type.
            value: Value.
        """
        location = ConstRds.REG_LOCATION

        with wrg.CreateKey(location, root_path) as reg:
            wrg.SetValueEx(reg, name, ConstRds.RESERVED, type_, value)

    def install_rds_role(self):
        """Configuring rds host.

        Raises:
            Exception: Shell error or execution failed.
        """
        rds_role_command = (
            'Get-WindowsFeature | ? { $_.Name -match "RDS-RD-Server" } | Install-WindowsFeature'
        )
        output = self.powershell_request(rds_role_command)
        if output.returncode != 0:
            text_error = output.stdout.strip()
            raise Exception(text_error)
        is_success = output.stdout.strip().split("\n")[0].split()[0].lower()
        is_true = output.stdout.strip().split("\n")[2].split()[0]
        if is_success != "success" or is_true != "True":
            raise Exception("Установка роли не удалась!")

    def configurate_registry(self):
        """Configuring the Windows registry
        to activate the Terminal Access service
        and disable app filter"""
        terminal_branch = r"SYSTEM\CurrentControlSet\Control\Terminal Server"
        filter_branch = r"SOFTWARE\Microsoft\Windows NT\CurrentVersion\Terminal Server\TSAppAllowList"
        self.__create_regedit_key(
            terminal_branch, "fDenyTSConnections", wrg.REG_DWORD, ConstRds.REG_DWORD_VALUES["Disable"]
        )
        self.__create_regedit_key(
            terminal_branch, "fSingleSessionPerUser", wrg.REG_DWORD, ConstRds.REG_DWORD_VALUES["Disable"]
        )
        self.__create_regedit_key(
            filter_branch, "fDisabledAllowList", wrg.REG_DWORD, ConstRds.REG_DWORD_VALUES["Enable"]
        )

    def configurate_firewall(self):
        """Configuring the Windows firewall.

        Raises:
            Exception: Shell error.
        """
        ssa_port = config.AGENT.LISTEN_PORT

        rds_command = f'netsh advfirewall firewall set rule group="{self.rds_lang_modifier}" new enable=yes'
        ssa_command = (
            "New-NetFirewallRule -DisplayName AllowSSA -Direction "
            f"Inbound -Protocol TCP -LocalPort {ssa_port} "
            "-Action Allow"
        )

        psh_commands = [rds_command, ssa_command]

        for command in psh_commands:
            output = self.powershell_request(command)
            if output.returncode != 0:
                text_error = output.stdout.strip()
                raise Exception(text_error)

    def configurate_all(self):
        """Configuring registry and firewall"""
        self.install_rds_role()
        self.configurate_registry()
        self.configurate_firewall()
