import logging
import uuid
from datetime import datetime, timedelta
from typing import Dict, Optional

# Fixme: these imports are made this way to make pyinstaller work.
#  Need to find another way for import migration in pyinstaller.
import jose
import jose.jwt as jwt
from cachetools import TLRUCache
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.types import CertificatePublicKeyTypes
from fastapi import Security
from fastapi.exceptions import HTTPException
from fastapi.openapi.models import HTTPBase as HTTPBaseModel
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from fastapi.security.base import SecurityBase
from fastapi.security.utils import get_authorization_scheme_param
from starlette.requests import Request
from starlette.status import HTTP_401_UNAUTHORIZED

import blowfish
from core.settings import config
from environment.checks import PlatformEnvironment
from excepts import AuthenticationError, DecryptionError, SSLCertError

if PlatformEnvironment.is_linux():
    from constants.linux.constants import ConstLinuxEnv as const
elif PlatformEnvironment.is_windows():
    from constants.windows.constants import ConstWindowsEnv as const

logger = logging.getLogger(__name__)


def get_public_key(path=config.SSL.CERTIFICATE_PATH) -> CertificatePublicKeyTypes:
    """Get public key.

    Reads the certificate file. The certificate must exist.

    Args:
        path: Path to the public key. Defaults to config.SSL.CERTIFICATE_PATH.

    Returns:
        CertificatePublicKeyTypes: Public key.

    Raises:
        SSLCertError: Error when file is not found, permission is denied
            or data in the certificate file is invalid
    """
    try:
        with open(path, "rb") as file:
            certificate = file.read()

        cert_obj = x509.load_pem_x509_certificate(certificate, default_backend())
        return cert_obj.public_key()
    except (PermissionError, FileNotFoundError) as ex:
        raise SSLCertError("Error retrieving public key") from ex

    except ValueError as ex:
        raise SSLCertError("Invalid certificate") from ex


def get_private_key(path=config.SSL.PRIVATE_KEY_PATH) -> bytes:
    """Get private key.

    Reads the private key file. The key must exist.

    Args:
        path: Path to the private key. Defaults to config.SSL.PRIVATE_KEY_PATH.

    Returns:
        bytes: Private key.

    Raises:
        SSLCertError: Error when file is not found, permission is denied
            or data in the private key file is invalid.
    """
    try:
        with open(path, "rb") as file:
            private_key = file.read()
            # Validation the private key
            serialization.load_pem_private_key(private_key, password=None, backend=default_backend())
            return private_key
    except (PermissionError, FileNotFoundError) as ex:
        raise SSLCertError("Error retrieving private key") from ex

    except ValueError as ex:
        raise SSLCertError("Invalid private key") from ex


def custom_ttu(_key, value, now):
    """Checking the TOKEN_TITLE and assigning the right time for the TOKEN_TITLE.

    ttu is time-to-use for TLRUCache.

    Args:
        _key: Cache key.
        value: Cache value.
        now: Time to create cache.

    Raises:
        Exception: Not correct create token ttu.

    Returns:
        datetime.datetime: Time the token will finish working.
    """
    if value == const.ACCESS_TOKEN_TITLE:
        return now + timedelta(seconds=const.ACCESS_TOKEN_TTL)
    elif value == const.REFRESH_TOKEN_TITLE:
        return now + timedelta(seconds=const.REFRESH_TOKEN_TTL)
    elif value == const.OTP_TOKEN_TITLE:
        return now + timedelta(seconds=const.OTP_TOKEN_TTL)
    else:
        raise Exception("Not correct create token ttu.")


def decrypt_password(password_encrypted: str, server_id: str) -> str:
    """
    Decrypting a password using a key.

    Args:
        password_encrypted: Encrypted password.
        server_id: Pre-shared key.

    Raises:
        DecryptionError: The error that occurs during decrypt.

    Returns:
        str: Decrypted password.
    """

    def unpad(value: bytes) -> bytes:
        len_value = len(value) - 1
        result_ord = -ord(value[len_value:])
        return value[:result_ord]

    try:
        splitted = password_encrypted.split(":")
        if len(splitted) != 1:
            type_crypt, value = splitted
            if type_crypt == "blowfish":
                if not server_id:
                    raise DecryptionError("Server ID is empty!")

                key = server_id.replace("-", "").encode("utf-8")
                cipher = blowfish.Cipher(key)
                bytes_password = bytes.fromhex(value)
                return unpad(b"".join(cipher.decrypt_ecb(bytes_password))).decode("utf-8")
            elif type_crypt in ("clear", "plain"):
                return value
            else:
                msg = f"Unknown еncryption type: {type_crypt}"
                raise DecryptionError(msg)
        else:
            return password_encrypted
    except ValueError as ex:
        raise DecryptionError("Decryption failed") from ex


class HTTPToken(SecurityBase):
    def __init__(
        self,
        *,
        scheme_name: Optional[str] = None,
        description: Optional[str] = None,
        auto_error: bool = False,
    ):
        self.model = HTTPBaseModel(scheme="token", description=description)
        self.scheme_name = scheme_name or self.__class__.__name__
        self.auto_error = auto_error

    async def __call__(self, request: Request) -> Optional[HTTPAuthorizationCredentials]:
        authorization: str = request.headers.get("Authorization")
        scheme, credentials = get_authorization_scheme_param(authorization)
        if not (authorization and scheme and credentials):
            if self.auto_error:
                raise HTTPException(status_code=HTTP_401_UNAUTHORIZED, detail="Not authenticated")
            return None
        if scheme.lower() != "token":
            if self.auto_error:
                raise HTTPException(
                    status_code=HTTP_401_UNAUTHORIZED,
                    detail="Invalid authentication credentials",
                )
            return None
        return HTTPAuthorizationCredentials(scheme=scheme, credentials=credentials)


class HealthAuthorization:
    security = HTTPToken()

    def verify(
        self,
        token_config: str,
        auth: HTTPAuthorizationCredentials = Security(security),
    ) -> bool:
        """
        Validate healthcheck token

        Args:
            auth (HTTPAuthorizationCredentials): credential of auth. Defaults to Security(security).
            token_config (str): token from config

        Returns:
            bool: answer if validate is ok
        """
        token = auth.credentials
        if token != token_config:
            return False
        return True


class BaseAuthorization:
    """
    Base JWT authentication mechanism.
    """

    def __init__(self):
        self._private_key = b""
        self._public_key = None
        self.algorithm = "RS256"
        self.token_whitelist = TLRUCache(maxsize=const.CACHE_LIMIT, ttu=custom_ttu, timer=datetime.now)

    @property
    def private_key(self) -> bytes:
        """The method retrieves the contents of the private key.

        Returns:
            bytes: Private key.
        """
        if not self._private_key:
            self._private_key = get_private_key()
        return self._private_key

    @property
    def public_key(self) -> CertificatePublicKeyTypes:
        """The method retrieves the contents of certificate.

        Returns:
            CertificatePublicKeyTypes: Public key.
        """
        if not self._public_key:
            self._public_key = get_public_key()
        return self._public_key

    def cache_verification(self, token) -> bool:
        """The method that checks the token in the allowed cache list.

        Args:
            token: Refresh or access token.

        Returns:
            bool: The token in whitelist.
        """
        return token in self.token_whitelist  # Checking that the token has not been revoked


class JWTAuthorization(BaseAuthorization):
    """
    JWT authentication mechanism

    Handled errors:
    - AuthenticationError (module excepts): the error that occurs
    during authentication
    - JWTError: called in case of validation error jwt token
    """

    security = HTTPBearer()

    def refresh(self, refresh_token) -> Dict[str, str]:
        """
        A method that updates the refresh token and returns a new pair of access, refresh tokens.

        Args:
            refresh_token: Refresh token data.

        Raises:
            AuthenticationError: Invalid refresh token or expired token.

        Returns:
            Dict[str, str]: Authentication token data.
        """
        try:
            payload = jwt.decode(refresh_token, self.public_key, algorithms=[self.algorithm])

            token_type = payload.get("typ")
            token_is_valid = self.cache_verification(refresh_token)
            if token_type != const.REFRESH_TOKEN_TITLE or not token_is_valid:
                raise AuthenticationError("Invalid refresh token")
            # The used refresh token must be revoked
            del self.token_whitelist[refresh_token]

            return self.create_tokens(payload)
        except jwt.ExpiredSignatureError as ex:
            raise AuthenticationError("Expired refresh token") from ex

        except jose.JWTError as ex:
            msg = f"Invalid refresh token: {str(ex)}"
            raise AuthenticationError(msg) from ex

    def verify(self, auth: HTTPAuthorizationCredentials = Security(security)) -> dict:
        """The method that validates the token in the request handler.

        Args:
            auth: Access token. Defaults to Security(security).

        Raises:
            AuthenticationError: Token is not valid or jwt validation error.

        Returns:
            dict: Authentication jwt data.
        """
        try:
            token = auth.credentials
            payload = jwt.decode(token, self.public_key, algorithms=[self.algorithm])
            token_is_valid = self.cache_verification(token)

            if not token_is_valid:
                raise AuthenticationError("Token is not valid.")

        except jose.JWTError as ex:
            msg = f"JWT validation error: {str(ex)}"
            raise AuthenticationError(msg) from ex

        return payload

    def generate_tokens(self, payload: dict, ttl: int, typ: str) -> str:
        """Generate token.

        A method that provides generation of a new token including payload, lifetime and token type.

        Args:
            payload: Authentication token payload.
            ttl: Token ttl, lifetime.
            typ: Token type.

        Returns:
            str: Encode token.
        """
        release = datetime.utcnow()
        expire = release + timedelta(seconds=ttl)
        token_payload = payload.copy()
        token_payload.update({"exp": expire, "iat": release, "typ": typ})
        headers = {"alg": self.algorithm}
        token = jwt.encode(token_payload, self.private_key, algorithm=self.algorithm, headers=headers)

        self.token_whitelist[token] = typ

        return token

    def create_tokens(self, payload: dict) -> Dict[str, str]:
        """Method for creating a new pair of access, refresh tokens.

        Args:
            payload: Authentication token payload.

        Returns:
            Dict[str, str]: Access and refresh token.
        """
        access_token = self.generate_tokens(payload, const.ACCESS_TOKEN_TTL, const.ACCESS_TOKEN_TITLE)
        refresh_token = self.generate_tokens(payload, const.REFRESH_TOKEN_TTL, const.REFRESH_TOKEN_TITLE)

        return {"access_token": access_token, "refresh_token": refresh_token}


class OTPAuthorization(BaseAuthorization):
    """
    OTP authentication mechanism

    Handled errors:
    - AuthenticationError (module excepts): the error that occurs
    during authentication
    - JWTError: called in case of validation error OTP token
    """

    def verify(self, token: str) -> None:
        """The method that validates the OTP token.

        Args:
            token: OTP token.

        Raises:
            AuthenticationError: Token is not valid or validation error.
        """
        try:
            jwt.decode(token, self.public_key, algorithms=[self.algorithm])
            token_is_valid = self.cache_verification(token)

            if not token_is_valid:
                raise AuthenticationError("Token is not valid.")

            del self.token_whitelist[token]

        except jose.JWTError as ex:
            msg = f"Validation error: {str(ex)}"
            raise AuthenticationError(msg) from ex

    def generate_tokens(self, typ: str) -> str:
        """Method for generating a new OTP token.

        Args:
            typ: Token type.

        Returns:
            str: OTP token.
        """

        release = datetime.utcnow()
        sub = uuid.uuid4().hex  # makes the token unique
        token_payload = {"iat": release, "typ": typ, "sub": sub}
        headers = {"alg": self.algorithm}
        token = jwt.encode(token_payload, self.private_key, algorithm=self.algorithm, headers=headers)

        self.token_whitelist[token] = typ

        return token

    def create_tokens(self) -> str:
        """Method for creating a new OTP token.

        Returns:
            str: OTP token.
        """
        return self.generate_tokens(const.OTP_TOKEN_TITLE)
